class CacheError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NotInitError(CacheError):
    def __init__(self):
        super().__init__("the cache should be inited before using")


class NotFoundStoreError(CacheError):
    def __init__(self, store_type, current_type_name):
        super().__init__(f"Unsupported ${store_type}: {current_type_name}")


class ParamError(CacheError):
    def __init__(self, message):
        super().__init__(message)



from abc import ABCMeta, abstractmethod


class SimilarityEvaluation(metaclass=ABCMeta):
    @abstractmethod
    def evaluation(self, src_dict, cache_dict, **kwargs): pass

    @abstractmethod
    def range(self): pass



from gptcache.utils import import_onnxruntime, import_huggingface_hub
from .similarity_evaluation import SimilarityEvaluation
import_onnxruntime()
import_huggingface_hub()
import numpy as np
import os
from transformers import AutoTokenizer
from pathlib import Path
from huggingface_hub import hf_hub_download
from typing import List
import onnxruntime

def pad_sequence(input_ids_list, padding_value=0):
    max_len = max(len(sequence) for sequence in input_ids_list)
    padded_sequences = np.full((len(input_ids_list), max_len), padding_value)
    for i, sequence in enumerate(input_ids_list):
        padded_sequences[i, :len(sequence)] = sequence
    return padded_sequences

class Onnx(SimilarityEvaluation):
    def __init__(self, model = 'GPTCache/albert-duplicate-onnx'):
        tokenizer_name = 'albert-base-v2'
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)
        self.model = model
        onnx_model_path = hf_hub_download(repo_id=model, filename = 'model.onnx')
        self.ort_session = onnxruntime.InferenceSession(onnx_model_path)

    # WARNING: the model cannot evaluate text with more than 512 tokens
    def evaluation(self, src_dict, cache_dict, **kwargs):
        try:
            src_question = src_dict["question"]
            cache_question = cache_dict["question"]
            if src_question == cache_question:
                return 1
            return self.inference(src_question, [cache_question])
        except Exception:
            return 0

    def range(self):
        return 0.0, 1.0

    def inference(self, reference: str, candidates: List[str]) -> np.ndarray:
 
         n_candidates= len(candidates)
         inference_texts = [{'text_a': reference, 'text_b': candidate } for candidate in  candidates ]
         batch_encoding_list = [self.tokenizer.encode_plus(text['text_a'], text['text_b'], padding='longest') for text in inference_texts]
 
         input_ids_list = [np.array(encode.input_ids) for encode in batch_encoding_list] 
         attention_mask_list = [np.array(encode.attention_mask) for encode in batch_encoding_list] 
         token_type_ids_list = [np.array(encode.token_type_ids) for encode in batch_encoding_list]
 
         padded_input_ids = pad_sequence(input_ids_list, padding_value=self.tokenizer.pad_token_id)
         padded_attention_mask = pad_sequence(attention_mask_list, padding_value=self.tokenizer.pad_token_id)
         padded_token_type_ids = pad_sequence(token_type_ids_list, padding_value=self.tokenizer.pad_token_id)
 
         ort_inputs = {'input_ids': padded_input_ids.reshape(n_candidates,-1), 
                       'attention_mask': padded_attention_mask.reshape(n_candidates,-1),
                       'token_type_ids': padded_token_type_ids.reshape(n_candidates,-1)}
         ort_outputs = self.ort_session.run(None, ort_inputs)
         scores = ort_outputs[0][:,1]
         return scores



from gptcache.utils import import_sqlalchemy
import_sqlalchemy()

import numpy as np
from datetime import datetime
from sqlalchemy import func, create_engine, Column, Sequence
from sqlalchemy.types import String, DateTime, LargeBinary, Integer
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

from .base import ScalarStorage, TABLE_NAME, TABLE_NAME_SEQ

Base = declarative_base()


class CacheTable(Base):
    """
    cache_table
    """
    __tablename__ = TABLE_NAME

    uid = Column(Integer, primary_key=True, autoincrement=True)
    id = Column(String(500), nullable=False)
    data = Column(String(1000), nullable=False)
    reply = Column(String(1000), nullable=False)
    create_on = Column(DateTime, default=datetime.now)
    last_access = Column(DateTime, default=datetime.now)
    embedding_data = Column(LargeBinary, nullable=True)
    state = Column(Integer, default=0)


class CacheTableSequence(Base):
    """
    cache_table_sequence
    """
    __tablename__ = TABLE_NAME_SEQ

    uid = Column(Integer, Sequence('id_seq', start=1), primary_key=True, autoincrement=True)
    id = Column(String(500), nullable=False)
    data = Column(String(1000), nullable=False)
    reply = Column(String(1000), nullable=False)
    create_on = Column(DateTime, default=datetime.now)
    last_access = Column(DateTime, default=datetime.now)
    embedding_data = Column(LargeBinary, nullable=True)
    state = Column(Integer, default=0)


class SQLDataBase(ScalarStorage):
    """
    Using sqlalchemy to manage SQLite, PostgreSQL, MySQL, MariaDB, SQL Server and Oracle.
    """
    def __init__(self, url: str = 'sqlite:///./gpt_cache.db', db_type: str = 'sqlite'):
        self._url = url
        self._engine = None
        self._session = None
        self._db_type = db_type
        if self._db_type == 'oracle':
            self._model = CacheTableSequence
        else:
            self._model = CacheTable
        self.init()

    def init(self):
        self._engine = create_engine(self._url)
        Session = sessionmaker(bind=self._engine)
        self._session = Session()
        self.create()

    def create(self):
        self._model.__table__.create(bind=self._engine, checkfirst=True)

    def insert(self, key, data, reply, embedding_data: np.ndarray = None):
        embedding_data = embedding_data.tobytes()
        model_obj = self._model(id=key, data=data, reply=reply, embedding_data=embedding_data)
        self._session.add(model_obj)
        self._session.commit()

    def get_data_by_ids(self, keys):
        res = self._session.query(self._model.data, self._model.reply).filter(self._model.id.in_(keys)).filter(self._model.state == 0).all()
        return res

    def get_data_by_id(self, key):
        res = self._session.query(self._model.data, self._model.reply).filter(self._model.id == key).filter(self._model.state == 0).first()
        return res

    def get_ids_by_state(self, state):
        res = self._session.query(self._model.id).filter(self._model.state == state).all()
        return res

    def get_embedding_data(self, offset, size):
        res = self._session.query(self._model.embedding_data).order_by(self._model.uid.asc()).limit(size).offset(offset).all()
        return res

    def update_access_time(self, key):
        self._session.query(self._model).filter(self._model.id == key).update({'last_access': datetime.now()})
        self._session.commit()

    def get_old_access(self, count):
        res = self._session.query(self._model.id).order_by(self._model.last_access.asc()).limit(count).all()
        return res

    def update_state(self, keys, state: int = -1):
        self._session.query(self._model).filter(self._model.id.in_(keys)).update({'state': state})
        self._session.commit()

    def remove_by_state(self):
        res = self._session.query(self._model).filter(self._model.state == -1).delete()
        self._session.commit()
        return res

    def count(self, state: int = 0, is_all: bool = False):
        if is_all:
            return self._session.query(func.count(self._model.id)).scalar()
        return self._session.query(func.count(self._model.id)).filter(self._model.state == state).scalar()

    def close(self):
        self._session.close()



import numpy as np


class EvictionManager:
    MAX_MARK_COUNT = 5000
    MAX_MARK_RATE = 0.1
    BATCH_SIZE = 100000

    def __init__(self, scalar_storage, vector_base):
        self._scalar_storage = scalar_storage
        self._vector_base = vector_base

    def check_evict(self):
        mark_count = self._scalar_storage.count(state=-1)
        all_count = self._scalar_storage.count(is_all=True)
        if mark_count > self.MAX_MARK_COUNT or mark_count / all_count > self.MAX_MARK_RATE:
            return True
        return False

    def delete(self):
        mark_ids = self._scalar_storage.get_ids_by_state(state=-1)
        mark_ids = [i[0] for i in mark_ids]
        self._scalar_storage.remove_by_state()
        self._vector_base.delete(mark_ids)

    def rebuild(self):
        self._scalar_storage.remove_by_state()
        count = self._scalar_storage.count()
        offset = 0
        while offset < count:
            data = self._scalar_storage.get_embedding_data(offset, self.BATCH_SIZE)
            np_data = [np.frombuffer(d[0], np.float32) for d in data]
            self._vector_base.rebuild(np_data)
            offset += self.BATCH_SIZE

    def soft_evict(self, count):
        marked_keys = self._scalar_storage.get_old_access(count)
        marked_keys = [i[0] for i in marked_keys]
        self._scalar_storage.update_state(marked_keys)



from abc import ABCMeta, abstractmethod

import numpy as np

TABLE_NAME = 'cache_table'
TABLE_NAME_SEQ = 'cache_table_sequence'


class ScalarStorage(metaclass=ABCMeta):
    """
    BaseStorage for scalar data.
    """
    @abstractmethod
    def init(self, **kwargs): pass

    @abstractmethod
    def create(self, **kwargs): pass

    @abstractmethod
    def insert(self, key, data, reply, embedding_data: np.ndarray = None): pass

    @abstractmethod
    def get_data_by_id(self, key): pass

    @abstractmethod
    def get_embedding_data(self, limit, offset): pass

    @abstractmethod
    def remove_by_state(self): pass

    @abstractmethod
    def update_access_time(self, keys): pass

    @abstractmethod
    def update_state(self, keys): pass

    @abstractmethod
    def count(self): pass

    @abstractmethod
    def close(self): pass



import atexit
import os
import time

import openai
from .embedding.string import to_embeddings as string_embedding
from .cache.data_manager import DataManager
from .cache.factory import get_data_manager
from .processor.post import first
from .processor.pre import last_content
from .similarity_evaluation.similarity_evaluation import SimilarityEvaluation
from .similarity_evaluation.string import AbsoluteEvaluation
from .utils.error import CacheError


def cache_all(*args, **kwargs):
    return True


def time_cal(func, func_name=None, report_func=None):
    def inner(*args, **kwargs):
        time_start = time.time()
        res = func(*args, **kwargs)
        delta_time = time.time() - time_start
        if cache.config.log_time_func:
            cache.config.log_time_func(func.__name__ if func_name is None else func_name, delta_time)
        if report_func is not None:
            report_func(delta_time)
        return res

    return inner


class Config:

    def __init__(self,
                 log_time_func=None,
                 similarity_threshold=0.8,
                 ):
        if similarity_threshold < 0 or similarity_threshold > 1:
            raise CacheError("Invalid the similarity threshold param")
        self.log_time_func = log_time_func
        self.similarity_threshold = similarity_threshold


class Report:
    def __init__(self):
        self.embedding_all_time = 0
        self.embedding_count = 0
        self.search_all_time = 0
        self.search_count = 0
        self.hint_cache_count = 0

    def embedding(self, delta_time):
        self.embedding_all_time += delta_time
        self.embedding_count += 1

    def search(self, delta_time):
        self.search_all_time += delta_time
        self.search_count += 1

    def average_embedding_time(self):
        return round(self.embedding_all_time / self.embedding_count if self.embedding_count != 0 else 0, 4)

    def average_search_time(self):
        return round(self.search_all_time / self.search_count if self.embedding_count != 0 else 0, 4)

    def hint_cache(self):
        self.hint_cache_count += 1


class Cache:
    similarity_evaluation: SimilarityEvaluation

    # it should be called when start the cache system
    def __init__(self):
        self.has_init = False
        self.cache_enable_func = None
        self.pre_embedding_func = None
        self.embedding_func = None
        self.data_manager = None
        self.post_process_messages_func = None
        self.config = Config()
        self.report = Report()
        self.next_cache = None

    def init(self,
             cache_enable_func=cache_all,
             pre_embedding_func=last_content,
             embedding_func=string_embedding,
             data_manager: DataManager = get_data_manager("map"),
             similarity_evaluation=AbsoluteEvaluation(),
             post_process_messages_func=first,
             config=Config(),
             next_cache=None,
             **kwargs
             ):
        self.has_init = True
        self.cache_enable_func = cache_enable_func
        self.pre_embedding_func = pre_embedding_func
        self.embedding_func = embedding_func
        self.data_manager: DataManager = data_manager
        self.similarity_evaluation = similarity_evaluation
        self.post_process_messages_func = post_process_messages_func
        self.data_manager.init(**kwargs)
        self.config = config
        self.next_cache = next_cache

        @atexit.register
        def close():
            try:
                self.data_manager.close()
            except Exception as e:
                print(e)

    @staticmethod
    def set_openai_key():
        openai.api_key = os.getenv("OPENAI_API_KEY")


cache = Cache()



from .similarity_evaluation import SimilarityEvaluation


class SearchDistanceEvaluation(SimilarityEvaluation):

    def __init__(self, max_distance=4.0, positive=False):
        self.max_distance = max_distance
        self.positive = positive

    def evaluation(self, src_dict, cache_dict, **kwargs):
        distance, _ = cache_dict["search_result"]
        if distance < 0:
            distance = 0
        elif distance > self.max_distance:
            distance = self.max_distance
        if self.positive:
            return distance
        return self.max_distance - distance

    def range(self):
        return 0.0, self.max_distance



__all__ = ['import_pymilvus', 'import_huggingface_hub',
           'import_faiss', 'import_chromadb',
           'import_sqlalchemy', 'import_sql_client',
           'import_huggingface', 'import_torch',
           'import_sbert', 'import_onnxruntime', 'import_cohere' ]
from .dependency_control import prompt_install


# pylint: disable=unused-import
# pylint: disable=ungrouped-imports
# pragma: no cover
def import_pymilvus():
    try:
        import pymilvus
    except ModuleNotFoundError as e:
        prompt_install('pymilvus')
        import pymilvus

def import_sbert():
    try:
        import sentence_transformers
    except ModuleNotFoundError:
        prompt_install('sentence-transformers')
        import sentence_transformers

def import_cohere():
    try:
        import cohere
    except ModuleNotFoundError:
        prompt_install('cohere')
        import cohere

def import_huggingface():
    try:
        import transformers
    except ModuleNotFoundError as e:
        prompt_install('transformers')
        import transformers

def import_torch():
    try:
        import torch
    except ModuleNotFoundError:
        prompt_install('torch')
        import torch

def import_huggingface_hub(): 
    try:
        import huggingface_hub
    except ModuleNotFoundError as e:
        prompt_install('huggingface-hub')
        import huggingface_hub

def import_onnxruntime():
    try:
        import onnxruntime
    except ModuleNotFoundError as e:
        prompt_install('onnxruntime')
        import onnxruntime

def import_faiss():
    try:
        import faiss
    except ModuleNotFoundError as e:
        prompt_install('faiss-cpu==1.6.5')
        import faiss


def import_chromadb():
    try:
        import chromadb
    except ModuleNotFoundError as e:
        prompt_install('chromadb')
        import chromadb


def import_sqlalchemy():
    try:
        import sqlalchemy
    except ModuleNotFoundError as e:
        prompt_install('sqlalchemy')
        import sqlalchemy


def import_postgresql():
    try:
        import psycopg2
    except ModuleNotFoundError as e:
        prompt_install('psycopg2-binary')
        import psycopg2


def import_pymysql():
    try:
        import pymysql
    except ModuleNotFoundError as e:
        prompt_install('pymysql')
        import pymysql


# `brew install unixodbc` in mac
# and install PyODBC driver.
def import_pyodbc():
    try:
        import pyodbc
    except ModuleNotFoundError as e:
        prompt_install('pyodbc')
        import pyodbc


# install cx-Oracle driver.
def import_cxoracle():
    try:
        import cx_Oracle
    except ModuleNotFoundError as e:
        prompt_install('cx_Oracle')
        import cx_Oracle


def import_sql_client(db_name):
    if db_name == 'postgresql':
        import_postgresql()
    elif db_name in ['mysql', 'mariadb']:
        import_pymysql()
    elif db_name == 'sqlserver':
        import_pyodbc()
    elif db_name == 'oracle':
        import_cxoracle()



from gptcache.utils import import_pymilvus
import_pymilvus()

from uuid import uuid4
import numpy as np
from .base import VectorBase, ClearStrategy
from pymilvus import (
    connections,
    utility,
    FieldSchema,
    DataType,
    CollectionSchema,
    Collection,
    MilvusException
)


class Milvus(VectorBase):

    def __init__(self, **kwargs):
        host = kwargs.get("host", "localhost")
        port = kwargs.get("port", 19530)
        user = kwargs.get("user", "")
        password = kwargs.get("password", "")
        use_security = kwargs.get("is_https", False)

        self.default_search_params = {
            "IVF_FLAT": {"metric_type": "L2", "params": {"nprobe": 10}},
            "IVF_SQ8": {"metric_type": "L2", "params": {"nprobe": 10}},
            "IVF_PQ": {"metric_type": "L2", "params": {"nprobe": 10}},
            "HNSW": {"metric_type": "L2", "params": {"ef": 10}},
            "RHNSW_FLAT": {"metric_type": "L2", "params": {"ef": 10}},
            "RHNSW_SQ": {"metric_type": "L2", "params": {"ef": 10}},
            "RHNSW_PQ": {"metric_type": "L2", "params": {"ef": 10}},
            "IVF_HNSW": {"metric_type": "L2", "params": {"nprobe": 10, "ef": 10}},
            "ANNOY": {"metric_type": "L2", "params": {"search_k": 10}},
            "AUTOINDEX": {"metric_type": "L2", "params": {}},
        }

        try:
            i = [
                connections.get_connection_addr(x[0])
                for x in connections.list_connections()
            ].index({"host": host, "port": port})
            self.alias = connections.list_connections()[i][0]
        except ValueError:
            # Connect to the Milvus instance using the passed in Environment variables
            self.alias = uuid4().hex
            connections.connect(
                alias=self.alias,
                host=host,
                port=port,
                user=user,  # type: ignore
                password=password,  # type: ignore
                secure=use_security,
            )
        collection_name = kwargs.get("collection_name", "gptcache")
        create_new = kwargs.get("create_new", False)
        dim = kwargs.get("dim", 0)
        if dim <= 0:
            raise ValueError(f"invalid dim param `{dim}` in the Milvus vector store")
        self.dim = dim
        self.top_k = kwargs.get("top_k", 1)
        self.index_params = kwargs.get("index_params", None)
        self._create_collection(collection_name, create_new)

        self.search_params = (
                kwargs.get("search_params", None) or self.default_search_params[self.index_params["index_type"]]
        )

    def _create_collection(self, collection_name, create_new):
        if utility.has_collection(collection_name, using=self.alias) and create_new:
            utility.drop_collection(collection_name, using=self.alias)

        if utility.has_collection(collection_name, using=self.alias) is False:
            schema = [
                FieldSchema(name="pk", dtype=DataType.VARCHAR, is_primary=True, auto_id=False, max_length=65535),
                FieldSchema(name="embedding", dtype=DataType.FLOAT_VECTOR, dim=self.dim),
            ]
            schema = CollectionSchema(schema)
            self.col = Collection(
                collection_name,
                schema=schema,
                consistency_level="Strong",
                using=self.alias,
            )
        else:
            self.col = Collection(
                collection_name, consistency_level="Strong", using=self.alias
            )

        if len(self.col.indexes) == 0:
            if self.index_params is not None:
                self.col.create_index("embedding", index_params=self.index_params)
            else:
                try:
                    print("Attempting creation of Milvus default index")
                    i_p = {
                        "metric_type": "L2",
                        "index_type": "HNSW",
                        "params": {"M": 8, "efConstruction": 64},
                    }

                    self.col.create_index("embedding", index_params=i_p)
                    self.index_params = i_p
                    print("Creation of Milvus default index successful")
                except MilvusException as e:
                    print("milvus e", e)
                    print("Attempting creation of default index")
                    i_p = {"metric_type": "L2", "index_type": "AUTOINDEX", "params": {}}
                    self.col.create_index("embedding", index_params=i_p)
                    self.index_params = i_p
        else:
            self.index_params = self.col.indexes[0].to_dict()["index_param"]

        self.col.load()

    def add(self, key: str, data: np.ndarray):
        entities = [
            [key],
            data.reshape(1, self.dim)
        ]
        self.col.insert(entities)

    def search(self, data: np.ndarray):
        search_result = self.col.search(
            data=data.reshape(1, -1).tolist(),
            anns_field="embedding",
            param=self.search_params,
            limit=self.top_k,
            output_fields=[
                "pk"
            ]
        )
        pks = {}
        for hit in search_result[0]:
            pks[hit.entity.get("pk")] = hit.score

        query_pks = ",".join(['"' + x + '"' for x in pks])
        query_result = self.col.query(expr=f"pk in [{query_pks}]", output_fields=["pk", "embedding"])

        search_tuples = []
        for query_row in query_result:
            search_tuples.append((pks[query_row["pk"]], np.array(query_row["embedding"])))
        return search_tuples

    def clear_strategy(self):
        return ClearStrategy.DELETE    

    def delete(self, ids):
        pks = ",".join(['"' + x + '"' for x in ids])
        self.col.delete(f"pk in [{pks}]")

    def close(self):
        self.col.flush(_async=True)



def last_content(data, **kwargs):
    return data.get("messages")[-1]["content"]


def all_content(data, **kwargs):
    s = ""
    messages = data.get("messages")
    for i, message in enumerate(messages):
        if i == len(messages) - 1:
            s += message["content"]
        else:
            s += message["content"] + "\n"
    return s


def nop(data, **kwargs):
    return data


def get_prompt(data, **kwargs):
    return data.get("prompt")



from abc import ABC, abstractmethod
from typing import List
from enum import Enum


class ClearStrategy(Enum):
    REBUILD = 0
    DELETE = 1


class VectorBase(ABC):

    def init(self, **kwargs):
        pass

    @abstractmethod
    def add(self, key: str, data: 'ndarray'):
        pass

    @abstractmethod
    def search(self, data: 'ndarray'):
        pass

    @abstractmethod
    def clear_strategy(self):
        pass    

    def rebuild(self) -> bool:
        raise NotImplementedError

    def delete(self, ids) -> bool:
        raise NotImplementedError

    @abstractmethod
    def close(self):
        pass



import logging
import copy
from ..core import cache, time_cal
from ..utils.error import NotInitError


def adapt(llm_handler, cache_data_convert, update_cache_callback, *args, **kwargs):
    copy_kwargs = copy.deepcopy(kwargs)
    chat_cache = kwargs.pop("cache_obj", cache)
    if not chat_cache.has_init:
        raise NotInitError()
    cache_enable = chat_cache.cache_enable_func(*args, **kwargs)
    context = kwargs.pop("cache_context", {})
    embedding_data = None
    # you want to retry to send the request to chatgpt when the cache is negative
    cache_skip = kwargs.pop("cache_skip", False)
    pre_embedding_data = chat_cache.pre_embedding_func(kwargs,
                                                       extra_param=context.get("pre_embedding_func", None))
    if cache_enable and not cache_skip:
        embedding_data = time_cal(chat_cache.embedding_func,
                                  func_name="embedding",
                                  report_func=chat_cache.report.embedding,
                                  )(pre_embedding_data, extra_param=context.get("embedding_func", None))
        cache_data_list = time_cal(chat_cache.data_manager.search,
                                   func_name="search",
                                   report_func=chat_cache.report.search,
                                   )(embedding_data, extra_param=context.get('search_func', None))
        if cache_data_list is None:
            cache_data_list = []
        cache_answers = []
        similarity_threshold = chat_cache.config.similarity_threshold
        min_rank, max_rank = chat_cache.similarity_evaluation.range()
        rank_threshold = (max_rank - min_rank) * similarity_threshold
        for cache_data in cache_data_list:
            ret = chat_cache.data_manager.get_scalar_data(
                cache_data, extra_param=context.get('get_scalar_data', None))
            if ret is None:
                continue
            cache_question, cache_answer = ret
            rank = chat_cache.similarity_evaluation.evaluation({
                    "question": pre_embedding_data,
                    "embedding": embedding_data,
                }, {
                    "question": cache_question,
                    "answer": cache_answer,
                    "search_result": cache_data,
                }, extra_param=context.get('evaluation_func', None))
            if rank_threshold <= rank:
                cache_answers.append((rank, cache_answer))
        cache_answers = sorted(cache_answers, key=lambda x: x[0], reverse=True)
        if len(cache_answers) != 0:
            return_message = chat_cache.post_process_messages_func([t[1] for t in cache_answers])
            chat_cache.report.hint_cache()
            return cache_data_convert(return_message)

    next_cache = chat_cache.next_cache
    if next_cache:
        kwargs = copy_kwargs
        kwargs["cache_obj"] = next_cache
        llm_data = adapt(llm_handler, cache_data_convert, update_cache_callback, *args, **kwargs)
    else:
        llm_data = llm_handler(*args, **kwargs)

    if cache_enable:
        try:
            def update_cache_func(handled_llm_data):
                chat_cache.data_manager.save(pre_embedding_data,
                                             handled_llm_data,
                                             embedding_data,
                                             extra_param=context.get('save_func', None))

            llm_data = update_cache_callback(llm_data, update_cache_func)
        except Exception as e:
            logging.warning(f"failed to save the data to cache, error:{e}")
    return llm_data



from gptcache.utils import import_huggingface, import_torch
import_huggingface()
import_torch()

import numpy as np
import torch
from transformers import AutoTokenizer, AutoModel


class Huggingface:
    """Generate sentence embedding for given text using pretrained models from Huggingface transformers.

    :param model: model name, defaults to "sentence-transformers/all-MiniLM-L6-v2".
    :type model: str

    Example:
        .. code-block:: python
        
            from gptcache.embedding import Huggingface
            
            test_sentence = "Hello, world." 
            encoder = Huggingface(model="sentence-transformers/all-MiniLM-L6-v2")
            embed = encoder.to_embeddings(test_sentence)
    """
    def __init__(self, model: str="sentence-transformers/all-MiniLM-L6-v2"):
        self.model = AutoModel.from_pretrained(model)
        self.model.eval()

        self.tokenizer = AutoTokenizer.from_pretrained(model)
        if not self.tokenizer.pad_token:
            self.tokenizer.pad_token = '[PAD]'
        try:
            self.__dimension = self.model.config.hidden_size
        except Exception:
            from transformers import AutoConfig
            config = AutoConfig.from_pretrained(model)
            self.__dimension = config.hidden_size

    def to_embeddings(self, data, **kwargs):
        """Generate embedding given text input

        :param data: text in string.
        :type data: str

        :return: a text embedding in shape of (dim,).
        """
        if not isinstance(data, list):
            data = [data]
        inputs = self.tokenizer(data, padding=True, truncation=True, return_tensors='pt')
        outs = self.model(**inputs).last_hidden_state
        emb = self.post_proc(outs, inputs).squeeze(0).detach().numpy()
        return np.array(emb).astype('float32')
    
    def post_proc(self, token_embeddings, inputs):
        attention_mask = inputs['attention_mask']
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        sentence_embs = torch.sum(
            token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)
        return sentence_embs

    @property
    def dimension(self):
        """Embedding dimension.

        :return: embedding dimension
        """
        return self.__dimension


__all__ = ['Onnx']

from gptcache.utils.lazy_import import LazyImport

onnx = LazyImport('onnx', globals(), 'gptcache.similarity_evaluation.onnx')

def Onnx(model = 'GPTCache/albert-duplicate-onnx'):
    return onnx.Onnx(model)



from .similarity_evaluation import SimilarityEvaluation


class AbsoluteEvaluation(SimilarityEvaluation):

    def evaluation(self, src_dict, cache_dict, **kwargs):
        return 1 if cache_dict["question"] == src_dict["question"] else 0

    def range(self):
        return 0, 1



import subprocess


def prompt_install(package): # pragma: no cover
    """
    Function used to prompt user to install a package.
    """
    try:
        cmd = f"pip install {package}"
        subprocess.check_call(cmd, shell=True)
        print(f'{package} installed successfully!')
    except subprocess.CalledProcessError:
        print(f'Ran into error installing {package}.')



__all__ = ['OpenAI', 'Huggingface', 'SBERT', 'Cohere', 'Onnx']


from gptcache.utils.lazy_import import LazyImport

openai = LazyImport('openai', globals(), 'gptcache.embedding.openai')
huggingface = LazyImport('huggingface', globals(), 'gptcache.embedding.huggingface')
sbert = LazyImport('sbert', globals(), 'gptcache.embedding.sbert')
onnx = LazyImport('onnx', globals(), 'gptcache.embedding.onnx')
cohere = LazyImport('cohere', globals(), 'gptcache.embedding.cohere')


def Cohere(model="large", api_key=None):
    return cohere.Cohere(model, api_key)

def OpenAI(model="text-embedding-ada-002", api_key=None):
    return openai.OpenAI(model, api_key)

def Huggingface(model="sentence-transformers/all-mpnet-base-v2"):
    return huggingface.Huggingface(model)

def SBERT(model="all-MiniLM-L6-v2"):
    return sbert.SBERT(model)

def Onnx(model="GPTCache/paraphrase-albert-onnx"):
    return onnx.Onnx(model)



import numpy as np

from .similarity_evaluation import SimilarityEvaluation


class NumpyNormEvaluation(SimilarityEvaluation):

    def __init__(self, enable_normal: bool = False):
        self.enable_normal = enable_normal

    @staticmethod
    def normalize(vec):
        magnitude = np.linalg.norm(vec)
        normalized_v = vec / magnitude
        return normalized_v

    def evaluation(self, src_dict, cache_dict, **kwargs):
        src_embedding = self.normalize(src_dict["embedding"]) if self.enable_normal else src_dict["embedding"]
        _, cache_embedding = cache_dict["search_result"]
        cache_embedding = self.normalize(cache_embedding) if self.enable_normal \
            else cache_embedding
        return np.linalg.norm(src_embedding - cache_embedding)

    def range(self):
        return 0.0, 2.0



from gptcache.utils import import_chromadb
import_chromadb()

import chromadb
from .base import VectorBase, ClearStrategy


class Chromadb(VectorBase):
    def __init__(self, **kwargs):
        client_settings = kwargs.get("client_settings", None)
        persist_directory = kwargs.get("persist_directory", None)
        collection_name = kwargs.get('collection_name', 'gptcache')
        self.top_k = kwargs.get("top_k", 1)

        if client_settings:
            self._client_settings = client_settings
        else:
            self._client_settings = chromadb.config.Settings()
            if persist_directory is not None:
                self._client_settings = chromadb.config.Settings(
                    chroma_db_impl="duckdb+parquet", persist_directory=persist_directory
                )
        self._client = chromadb.Client(self._client_settings)
        self._persist_directory = persist_directory
        self._collection = self._client.get_or_create_collection(
            name=collection_name
        )

    def add(self, key, data):
      self._collection.add(
          embeddings=[data], ids=[key]
      )

    def search(self, data):
        if self._collection.count() == 0:
            return []

        results = self._collection.query(
            query_embeddings=[data], n_results=self.top_k,
            include=['distances', 'embeddings']
           )
        return list(zip(results['distances'][0], results['embeddings'][0]))

    def clear_strategy(self):
        return ClearStrategy.DELETE    

    def delete(self, ids):
        self._collection.delete(ids)

    def close(self):
        return True



from typing import Iterator

import openai

from .adapter import adapt


class ChatCompletion:
    @classmethod
    def create(cls, *args, **kwargs):
        def llm_handler(*llm_args, **llm_kwargs):
            return openai.ChatCompletion.create(*llm_args, **llm_kwargs)

        def cache_data_convert(cache_data):
            if kwargs.get("stream", False):
                return construct_stream_resp_from_cache(cache_data)
            return construct_resp_from_cache(cache_data)

        def update_cache_callback(llm_data, update_cache_func):
            if not isinstance(llm_data, Iterator):
                update_cache_func(get_message_from_openai_answer(llm_data))
                return llm_data
            else:
                def hook_openai_data(it):
                    total_answer = ""
                    for item in it:
                        total_answer += get_stream_message_from_openai_answer(item)
                        yield item
                    update_cache_func(total_answer)

                return hook_openai_data(llm_data)

        return adapt(llm_handler, cache_data_convert, update_cache_callback, *args, **kwargs)


def construct_resp_from_cache(return_message):
    return {
        "gptcache": True,
        "choices": [
            {
                "message": {
                    "role": "assistant",
                    "content": return_message
                },
                "finish_reason": "stop",
                "index": 0
            }
        ]
    }


def construct_stream_resp_from_cache(return_message):
    return [
        {
            "choices": [
                {
                    "delta": {
                        "role": "assistant"
                    },
                    "finish_reason": None,
                    "index": 0
                }
            ],
        },
        {
            "choices": [
                {
                    "delta": {
                        "content": return_message
                    },
                    "finish_reason": None,
                    "index": 0
                }
            ],
        },
        {
            "gptcache": True,
            "choices": [
                {
                  "delta": {},
                  "finish_reason": "stop",
                  "index": 0
                }
            ],
        }
    ]


def get_message_from_openai_answer(openai_data):
    return openai_data['choices'][0]['message']['content']


def get_stream_message_from_openai_answer(openai_data):
    return openai_data['choices'][0]['delta'].get('content', '')



from gptcache.utils import import_onnxruntime, import_huggingface_hub, import_huggingface
import_huggingface()
import_onnxruntime()
import_huggingface_hub()
import numpy as np
import os
from transformers import AutoTokenizer, AutoConfig
from pathlib import Path
from huggingface_hub import hf_hub_download
import onnxruntime

class Onnx:
    """Generate text embedding for given text using ONNX Model.

    Example:
        .. code-block:: python
        
            from gptcache.embedding import onnx
            
            test_sentence = "Hello, world." 
            encoder = onnx(model="GPTCache/paraphrase-albert-onnx")
            embed = encoder.to_embeddings(test_sentence)
    """
    def __init__(self, api_key, model= "GPTCache/paraphrase-albert-onnx", **kwargs):
        tokenizer_name = 'sentence-transformers/paraphrase-albert-small-v2'
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)
        self.model = model
        onnx_model_path = hf_hub_download(repo_id=model, filename = 'model.onnx')
        self.ort_session = onnxruntime.InferenceSession(onnx_model_path)
        config = AutoConfig.from_pretrained('sentence-transformers/paraphrase-albert-small-v2')
        self.__dimension = config.hidden_size

    
    def to_embeddings(self, data, **kwargs):
        encoded_text = self.tokenizer.encode_plus(data, padding='max_length')
        ort_inputs = {'input_ids': np.array(encoded_text['input_ids']).reshape(1,-1), 
               'attention_mask': np.array(encoded_text['attention_mask']).reshape(1,-1),
               'token_type_ids': np.array(encoded_text['token_type_ids']).reshape(1,-1)}

        ort_outputs = self.ort_session.run(None, ort_inputs)
        ort_feat = ort_outputs[0]
        emb = self.post_proc(ort_feat, ort_inputs['attention_mask'])
        return emb.flatten()

    def post_proc(self, token_embeddings, attention_mask): 
        token_embeddings = token_embeddings
        input_mask_expanded = np.expand_dims(attention_mask, -1).repeat(token_embeddings.shape[-1], -1).astype(float)
        sentence_embs = np.sum(token_embeddings * input_mask_expanded, 1) / np.maximum(input_mask_expanded.sum(1), 1e-9)
        return sentence_embs

    @property
    def dimension(self):
        return self.__dimension




import hashlib
from abc import abstractmethod, ABCMeta
import pickle
import cachetools
import numpy as np

from .scalar_data.base import ScalarStorage
from .vector_data.base import VectorBase, ClearStrategy
from .eviction import EvictionManager
from ..utils.error import CacheError


class DataManager(metaclass=ABCMeta):
    @abstractmethod
    def init(self, **kwargs): pass

    @abstractmethod
    def save(self, question, answer, embedding_data, **kwargs): pass

    # should return the tuple, (question, answer)
    @abstractmethod
    def get_scalar_data(self, vector_data, **kwargs): pass

    @abstractmethod
    def search(self, embedding_data, **kwargs): pass

    @abstractmethod
    def close(self): pass


class MapDataManager(DataManager):
    def __init__(self, data_path, max_size, get_data_container=None):
        if get_data_container is None:
            self.data = cachetools.LRUCache(max_size)
        else:
            self.data = get_data_container(max_size)
        self.data_path = data_path

    def init(self, **kwargs):
        try:
            f = open(self.data_path, 'rb')
            self.data = pickle.load(f)
            f.close()
        except FileNotFoundError:
            # print(f'File <${self.data_path}> is not found.')
            return
        except PermissionError:
            raise CacheError(f'You don\'t have permission to access this file <${self.data_path}>.')

    def save(self, question, answer, embedding_data, **kwargs):
        self.data[embedding_data] = (question, answer)

    def get_scalar_data(self, vector_data, **kwargs):
        return vector_data

    def search(self, embedding_data, **kwargs):
        try:
            return [self.data[embedding_data]]
        except KeyError:
            return []

    def close(self):
        try:
            f = open(self.data_path, 'wb')
            pickle.dump(self.data, f)
            f.close()
        except PermissionError:
            print(f'You don\'t have permission to access this file <${self.data_path}>.')


def sha_data(data):
    if isinstance(data, list):
        data = np.array(data)
    m = hashlib.sha1()
    m.update(data.astype('float32').tobytes())
    return m.hexdigest()


def normalize(vec):
    magnitude = np.linalg.norm(vec)
    normalized_v = vec / magnitude
    return normalized_v


class SSDataManager(DataManager):
    s: ScalarStorage
    v: VectorBase

    def __init__(self, max_size, clean_size, s, v):
        self.max_size = max_size
        self.cur_size = 0
        self.clean_size = clean_size
        self.s = s
        self.v = v
        self.eviction = EvictionManager(s, v)

    def init(self, **kwargs):
        self.s.init(**kwargs)
        self.v.init(**kwargs)
        self.cur_size = self.s.count()

    def _clear(self):
        self.eviction.soft_evict(self.clean_size)
        if not self.eviction.check_evict():
            pass
        elif self.v.clear_strategy() == ClearStrategy.DELETE:
            self.eviction.delete()
        elif self.v.clear_strategy() == ClearStrategy.REBUILD:
            self.eviction.rebuild()
        else:
            raise RuntimeError('Unknown clear strategy')
        self.cur_size = self.s.count()

    def save(self, question, answer, embedding_data, **kwargs):
        if self.cur_size >= self.max_size:
            self._clear()
        embedding_data = normalize(embedding_data)
        key = sha_data(embedding_data)
        self.s.insert(key, question, answer, embedding_data.astype('float32'))
        self.v.add(key, embedding_data)
        self.cur_size += 1

    def get_scalar_data(self, search_data, **kwargs):
        distance, vector_data = search_data
        key = sha_data(vector_data)
        self.s.update_access_time(key)
        return self.s.get_data_by_id(key)

    def search(self, embedding_data, **kwargs):
        embedding_data = normalize(embedding_data)
        return self.v.search(embedding_data)

    def close(self):
        self.s.close()
        self.v.close()



from typing import Optional, List
from .adapter import adapt


class LangChainLLMs:
    def __init__(self, llm):
        self._llm = llm

    def __call__(self, prompt: str, stop: Optional[List[str]] = None, **kwargs):
        # TODO handle the array result
        return adapt(self._llm, cache_data_convert, update_cache_callback, prompt=prompt, stop=stop, **kwargs)


def cache_data_convert(cache_data):
    return cache_data


def update_cache_callback(llm_data, update_cache_func):
    update_cache_func(llm_data)
    return llm_data



from .data_manager import DataManager, SSDataManager
from .scalar_data import SQLDataBase, SQL_URL
from .vector_data import Milvus, Faiss, Chromadb
from ..utils.error import NotFoundStoreError, ParamError
from ..utils import import_sql_client


def get_data_manager(data_manager_name: str, **kwargs) -> DataManager:
    if data_manager_name == "map":
        from .data_manager import MapDataManager

        return MapDataManager(kwargs.pop("data_path", "data_map.txt"),
                              kwargs.pop("max_size", 100),
                              kwargs.pop("get_data_container", None))
    elif data_manager_name == "scalar_vector":
        scalar_store = kwargs.pop("scalar_store", None)
        vector_store = kwargs.pop("vector_store", None)
        max_size = kwargs.pop("max_size", 1000)
        clean_size = kwargs.pop("clean_size", int(max_size * 0.2))
        if scalar_store is None or vector_store is None:
            raise ParamError(f"Missing scalar_store or vector_store parameter for scalar_vector")
        return SSDataManager(max_size, clean_size, scalar_store, vector_store)
    else:
        raise NotFoundStoreError("data manager", data_manager_name)


def _get_scalar_store(scalar_store: str, **kwargs):
    if scalar_store in ["sqlite", "postgresql", "mysql", "mariadb", "sqlserver", "oracle"]:
        sql_url = kwargs.pop("sql_url", SQL_URL[scalar_store])
        import_sql_client(scalar_store)
        store = SQLDataBase(url=sql_url, db_type=scalar_store)
    else:
        raise NotFoundStoreError("scalar store", scalar_store)
    return store


def _get_common_params(**kwargs):
    max_size = kwargs.pop("max_size", 1000)
    clean_size = kwargs.pop("clean_size", int(max_size * 0.2))
    top_k = kwargs.pop("top_k", 1)
    dimension = kwargs.pop("dimension", 0)
    return max_size, clean_size, dimension, top_k


def _check_dimension(dimension):
    if dimension <= 0:
        raise ParamError(f"the data manager should set the 'dimension' parameter greater than zero, "
                         f"current: {dimension}")


# scalar_store + vector_store
def get_ss_data_manager(scalar_store: str, vector_store: str, **kwargs):
    max_size, clean_size, dimension, top_k = _get_common_params(**kwargs)
    scalar = _get_scalar_store(scalar_store, **kwargs)
    if vector_store == "milvus":
        _check_dimension(dimension)
        vector = Milvus(dim=dimension, top_k=top_k, **kwargs)
    elif vector_store == "faiss":
        _check_dimension(dimension)
        index_path = kwargs.pop("index_path", "faiss.index")
        vector = Faiss(index_path, dimension, top_k)
    elif vector_store == "chromadb":
        vector = Chromadb(top_k=top_k, **kwargs)
    else:
        raise NotFoundStoreError("vector store", vector_store)
    return SSDataManager(max_size, clean_size, scalar, vector)



import setuptools
from setuptools import find_packages
from typing import List

with open("README.md", "r") as fh:
    long_description = fh.read()


def parse_requirements(file_name: str) -> List[str]:
    with open(file_name) as f:
        return [
            require.strip() for require in f
            if require.strip() and not require.startswith('#')
        ]


setuptools.setup(
    name="gptcache",
    packages=find_packages(),
    version="0.1.3",
    author="SimFG",
    author_email="bang.fu@zilliz.com",
    description="GPT Cache, a powerful caching library that can be used to speed up and lower the cost of chat "
                "applications that rely on the LLM service. GPT Cache works as a memcache for AIGC applications, "
                "similar to how Redis works for traditional applications.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=parse_requirements('requirements.txt'),
    url="https://github.com/zilliztech/gptcache",
    license='http://www.apache.org/licenses/LICENSE-2.0',
    python_requires='>=3.8.1',
)


from gptcache.utils import import_faiss
import_faiss()

import os
import faiss
from faiss import IndexHNSWFlat, Index
import numpy as np
from .base import VectorBase, ClearStrategy


class Faiss(VectorBase):
    index: Index

    def __init__(self, index_file_path, dimension, top_k, skip_file=False):
        self.index_file_path = index_file_path
        self.dimension = dimension
        self.index = IndexHNSWFlat(dimension, 32)
        self.top_k = top_k
        if os.path.isfile(index_file_path) and not skip_file:
            self.index = faiss.read_index(index_file_path)

    def add(self, key:str, data: 'ndarray'):
        np_data = np.array(data).astype('float32').reshape(1, -1)
        self.index.add(np_data)

    def _mult_add(self, datas):
        np_data = np.array(datas).astype('float32')
        self.index.add(np_data)

    def search(self, data: 'ndarray'):
        if self.index.ntotal == 0:
            return None
        np_data = np.array(data).astype('float32').reshape(1, -1)
        D, I = self.index.search(np_data, self.top_k)
        distances = []
        for d in D[:1].reshape(-1):
            distances.append(d)
        vector_datas = [self.index.reconstruct(int(i)) for i in I[:1].reshape(-1)]
        return zip(distances, vector_datas)

    def clear_strategy(self):
        return ClearStrategy.REBUILD

    def rebuild(self, all_data):
        f = Faiss(self.index_file_path, self.dimension, top_k=self.top_k, skip_file=True)
        f._mult_add(all_data)
        return f

    def close(self):
        faiss.write_index(self.index, self.index_file_path)



__all__ = ['SQLDataBase', 'SQL_URL']

from gptcache.utils.lazy_import import LazyImport

sql_database = LazyImport('milvus', globals(), 'gptcache.cache.scalar_data.sqlalchemy')


def SQLDataBase(**kwargs):
    return sql_database.SQLDataBase(**kwargs)


SQL_URL = {
        'sqlite': 'sqlite:///./sqlite.db',
        'postgresql': 'postgresql+psycopg2://postgres:123456@127.0.0.1:5432/postgres',
        'mysql': 'mysql+pymysql://root:123456@127.0.0.1:3306/mysql',
        'mariadb': 'mariadb+pymysql://root:123456@127.0.0.1:3307/mysql',
        'sqlserver': 'mssql+pyodbc://sa:Strongpsw_123@127.0.0.1:1434/msdb?driver=ODBC+Driver+17+for+SQL+Server',
        'oracle': 'oracle+cx_oracle://oracle:123456@127.0.0.1:1521/?service_name=helowin&encoding=UTF-8&nencoding=UTF-8',
}



__all__ = ['Milvus', 'Faiss', 'Chromadb']

from gptcache.utils.lazy_import import LazyImport

milvus = LazyImport('milvus', globals(), 'gptcache.cache.vector_data.milvus')
faiss = LazyImport('faiss', globals(), 'gptcache.cache.vector_data.faiss')
chroma = LazyImport('chroma', globals(), 'gptcache.cache.vector_data.chroma')


def Milvus(**kwargs):
    return milvus.Milvus(**kwargs)


def Faiss(index_file_path, dimension, top_k, skip_file=False):
    return faiss.Faiss(index_file_path, dimension, top_k, skip_file)


def Chromadb(**kwargs):
    return chroma.Chromadb(**kwargs)
